#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD.
# See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import os
import numpy as np
import logging
import json
from io import StringIO
from ase.cell import Cell
from ase.io import vasp

from nomad.units import ureg
from nomad.parsing.file_parser import TextParser, Quantity
from nomad.datamodel.metainfo.simulation.run import Run, Program
from nomad.datamodel.metainfo.simulation.calculation import (
    Calculation, Energy, EnergyEntry, Forces, ForcesEntry, Stress, StressEntry,
    Thermodynamics, Dos, DosValues, BandStructure, BandEnergies)
from nomad.datamodel.metainfo.simulation.method import Method
from nomad.datamodel.metainfo.simulation.system import System, Atoms
from nomad.datamodel.metainfo.workflow import (
    Workflow, Elastic, DebyeModel, Phonon, Thermodynamics as WorkflowThermodynamics)


from .metainfo import m_env


class AflowOutParser(TextParser):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def init_quantities(self):

        def str_to_property(val_in):
            val = val_in.split('=')
            return val[0].strip().replace(' ', '_').lower(), val[-1].split('//')[0].strip()

        self._quantities = [
            Quantity(
                'property',
                r'\n *\[(.+)\](.+?=.+)', str_operation=str_to_property, repeats=True),
            Quantity(
                'section',
                r'(\[.+?\]START[\s\S]+?\]STOP)',
                repeats=True, sub_parser=TextParser(quantities=[
                    Quantity(
                        'name',
                        r'\[(.+)\]START', str_operation=lambda x: x.lower(), dtype=str),
                    Quantity(
                        'key_value',
                        r'\n *([^#]\S+)=(\S+)', repeats=True),
                    Quantity(
                        'array',
                        rf'\n *(\d[\s\S]+?\d\s*)\[.+?STOP', dtype=np.dtype(np.float64))]))]

    def parse(self, key=None):
        super().parse(key)
        for property in self._results.get('property', []):
            self._results[property[0]] = property[1]
        for section in self._results.get('section', []):
            if section.key_value is not None:
                result = dict()
                for key, value in section.get('key_value', []):
                    result[key] = value
                self._results[section.name] = result
            elif section.array is not None:
                self._results[section.name] = section.array


class AflowInParser(AflowOutParser):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def init_quantities(self):
        super().init_quantities()
        self._quantities += [
            Quantity(
                'aflow_version',
                r'Stefano Curtarolo \- \(AFLOW V([\d\.]+)\)'),
            Quantity(
                'poscar',
                r'\[VASP_POSCAR_MODE_EXPLICIT\]START\s*([\s\S]+?)\[VASP_POSCAR_MODE_EXPLICIT\]STOP',
                str_operation=lambda x: x, convert=False, repeats=True),
            Quantity(
                'aflow_composition',
                r'\[AFLOW\] COMPOSITION=(\S+)', sub_parser=TextParser(quantities=[
                    Quantity('species', r'([A-Z][a-z]*)', repeats=True, dtype=str),
                    Quantity('composition', r'(\d+)\|', repeats=True, dtype=np.dtype(np.int32))
                ])
            )
        ] + [Quantity(
            module.lower(),
            r'\n *\[AFLOW\_%s\]CALC([\s\S]+?)\[AFLOW\] \*' % module, sub_parser=TextParser(quantities=[
                Quantity(
                    'parameters',
                    r'\[AFLOW\_%s\](.+?)=(\S+)' % module, repeats=True)])) for module in [
                        'AEL', 'AGL', 'APL', 'QHA', 'AAPL']]

    def parse(self, key=None):
        super().parse(key)

        if self.get('poscar') is not None and self._results.get('geometry') is None:
            try:
                atoms = vasp.read_vasp(StringIO(self._results['poscar'][-1]))
                self._results['cell'] = atoms.get_cell()
                self._results['geometry'] = atoms.get_cell().cellpar()
                composition = self._results['aflow_composition']
                self._results['species'] = composition.species
                self._results['composition'] = [int(c) for c in composition._results['composition']]
                self._results['positions_cartesian'] = atoms.get_positions()
            except Exception:
                pass

        if self._results.get('loop') is None:
            self._results['loop'] = [module for module in [
                'ael', 'agl', 'apl', 'qha', 'aapl'] if module in self._results]


class AFLOWParser:
    def __init__(self):
        self.ael_parser = AflowOutParser()
        self.agl_parser = AflowOutParser()
        self.apl_parser = AflowOutParser()
        self.aflowin_parser = AflowInParser()
        self._metainfo_env = m_env

        self._metainfo_map = {
            'stiffness_tensor': 'elastic_constants_matrix_second_order',
            'compliance_tensor': 'compliance_matrix_second_order',
            'poisson_ratio': 'poisson_ratio_hill',
            'bulk_modulus_vrh': 'bulk_modulus_hill',
            'shear_modulus_vrh': 'shear_modulus_hill',
            'youngs_modulus_vrh': 'Young_modulus_hill',
            'pughs_modulus_ratio': 'pugh_ratio_hill', 'applied_pressure': 'x_aflow_ael_applied_pressure',
            'average_external_pressure': 'x_aflow_ael_average_external_pressure'
        }

    def init_parser(self):
        if '.json' in self.filepath:
            self.aflow_data = json.load(open(self.filepath))
        else:
            self.aflowin_parser.mainfile = self.filepath
            self.aflow_data = self.aflowin_parser

    def get_aflow_file(self, filename):
        files = [f for f in os.listdir(self.maindir) if filename in f]
        if not files:
            files = ['']
        return os.path.join(self.maindir, files[0])

    def parse_structures(self, module):
        try:
            structures = json.load(open(os.path.join(
                self.maindir, '%s_energy_structures.json' % module))).get('%s_energy_structures' % module, [])
        except Exception:
            structures = []

        for structure in structures:
            sec_calc = self.archive.run[-1].m_create(Calculation)
            sec_thermo = sec_calc.m_create(Thermodynamics)
            if structure.get('energy') is not None:
                sec_calc.energy = Energy(total=EnergyEntry(value=structure.get('energy') * ureg.eV))
            if structure.get('pressure') is not None:
                sec_thermo.pressure = structure.get('pressure') * ureg.kbar
            if structure.get('stress_tensor') is not None:
                sec_calc.stress = Stress(total=StressEntry(value=structure.get('stress_tensor') * ureg.kbar))
            if structure.get('structure') is not None:
                sec_system = self.archive.run[-1].m_create(System)
                sec_system.atoms = Atoms()
                struc = structure.get('structure')
                sec_system.atoms.labels = [atom.get('name') for atom in struc.get('atoms', [])]
                sec_system.atoms.concentrations = [atom.get('occupancy') for atom in struc.get('atoms', [])]
                if struc.get('lattice') is not None:
                    sec_system.atoms.lattice_vectors = struc.get(
                        'lattice') * ureg.angstrom * struc.get('scale', 1)
                positions = [atom.get('position') for atom in struc.get('atoms', [])]
                if struc.get('coordinates_type', 'direct').lower().startswith('d'):
                    if sec_system.atoms.lattice_vectors is not None:
                        positions = np.dot(positions, sec_system.atoms.lattice_vectors)
                sec_system.atoms.positions = positions

    def parse_agl(self):
        sec_run = self.archive.m_create(Run)
        sec_run.program = Program(
            name='AFlow', version=self.aflow_data.get('aflow_version', 'unknown'))

        self.parse_structures('AGL')

        self.agl_parser.mainfile = self.get_aflow_file('aflow.agl.out')
        thermal_properties = self.agl_parser.get('agl_thermal_properties_temperature')
        if thermal_properties is None:
            return

        sec_workflow = self.archive.m_create(Workflow)
        sec_workflow.run_ref = sec_run
        sec_workflow.type = 'debye_model'
        sec_debye = sec_workflow.m_create(DebyeModel)
        sec_thermo = sec_workflow.m_create(WorkflowThermodynamics)

        thermal_properties = np.reshape(thermal_properties, (len(thermal_properties) // 9, 9))
        thermal_properties = np.transpose(thermal_properties)
        energies = self.agl_parser.get('agl_energies_temperature')
        energies = np.reshape(energies, (len(energies) // 9, 9))
        energies = np.transpose(energies)

        sec_thermo.temperature = thermal_properties[0] * ureg.K
        sec_thermo.gibbs_free_energy = energies[1] * ureg.eV
        sec_thermo.vibrational_free_energy = energies[2] * ureg.meV
        sec_thermo.vibrational_internal_energy = energies[3] * ureg.meV
        sec_thermo.vibrational_entropy = energies[4] * ureg.meV / ureg.K
        sec_thermo.heat_capacity_c_v = thermal_properties[4] * ureg.boltzmann_constant
        sec_thermo.heat_capacity_c_p = thermal_properties[5] * ureg.boltzmann_constant
        sec_debye.thermal_conductivity = thermal_properties[1] * ureg.watt / ureg.m * ureg.K
        sec_debye.debye_temperature = thermal_properties[2] * ureg.K
        sec_debye.gruneisen_parameter = thermal_properties[3]
        sec_debye.thermal_expansion = thermal_properties[6] / ureg.K
        sec_debye.bulk_modulus_static = thermal_properties[7] * ureg.GPa
        sec_debye.bulk_modulus_isothermal = thermal_properties[8] * ureg.GPa

    def parse_ael(self):
        sec_run = self.archive.m_create(Run)
        sec_run.program = Program(
            name='AFlow', version=self.aflow_data.get('aflow_version', 'unknown'))

        self.parse_structures('AEL')

        self.ael_parser.mainfile = self.get_aflow_file('aflow.ael.out')
        sec_workflow = self.archive.m_create(Workflow)
        sec_workflow.run_ref = sec_run
        sec_workflow.type = 'elastic'
        sec_elastic = sec_workflow.m_create(Elastic)
        sec_elastic.energy_stress_calculator = 'vasp'
        sec_elastic.calculation_method = 'stress'
        sec_elastic.elastic_constants_order = 2

        paths = [d for d in self.aflow_data.get('files', []) if d.startswith('ARUN.AEL')]
        deforms = np.array([d.split('_')[-2:] for d in paths], dtype=np.dtype(np.float64))
        strains = [d[1] for d in deforms if d[0] == 1]
        sec_elastic.n_deformations = int(max(np.transpose(deforms)[0]))
        sec_elastic.n_strains = len(strains)
        sec_elastic.strain_maximum = max(strains) - 1.0

        for key, val in self.ael_parser.get('ael_results', {}).items():
            key = key.replace('ael_', '')
            key = self._metainfo_map.get(key, key)
            if 'modulus' in key or 'pressure' in key:
                val = val * ureg.GPa
            elif 'speed' in key:
                val = val * (ureg.m / ureg.s)
            elif 'temperature' in key:
                val = val * ureg.K
            setattr(sec_elastic, key, val)

        if self.ael_parser.ael_stiffness_tensor is not None:
            sec_elastic.elastic_constants_matrix_second_order = np.reshape(
                self.ael_parser.ael_stiffness_tensor, (6, 6)) * ureg.GPa

        if self.ael_parser.ael_compliance_tensor is not None:
            sec_elastic.compliance_matrix_second_order = np.reshape(
                self.ael_parser.ael_compliance_tensor, (6, 6))

    def parse_apl(self):
        sec_run = self.archive.m_create(Run)
        sec_run.program = Program(
            name='AFlow', version=self.aflow_data.get('aflow_version', 'unknown'))
        sec_scc = sec_run.m_create(Calculation)

        try:
            dos = np.transpose(
                np.loadtxt(self.get_aflow_file('flow.apl.phonon_dos.out.xz')))
        except Exception:
            dos = None

        if dos is not None:
            sec_dos = sec_scc.m_create(Dos, Calculation.dos_phonon)
            sec_dos.energies = dos[2] * ureg.millielectron_volt
            sec_dos.total.append(DosValues(value=dos[3] * (1 / ureg.millielectron_volt)))

        try:
            kpoints = np.transpose(
                np.loadtxt(self.get_aflow_file('aflow.apl.hskpoints.out.xz')))
            n_kpoints = int(max(kpoints[3])) + 1
            kpoints = kpoints[:3]
            kpoints = np.reshape(kpoints, (3, len(kpoints[0]) // n_kpoints, n_kpoints))
            kpoints = np.transpose(kpoints, axes=(1, 2, 0))

            bandstructure = np.transpose(
                np.loadtxt(self.get_aflow_file('aflow.apl.phonon_dispersion.out.xz')))
            bandstructure = bandstructure[2:]
            bandstructure = np.reshape(bandstructure, (
                len(bandstructure), len(bandstructure[0]) // n_kpoints, n_kpoints))
            bandstructure = np.transpose(bandstructure, axes=(1, 2, 0))
        except Exception:
            kpoints = None

        if kpoints is not None:
            sec_bandstructure = sec_scc.m_create(BandStructure, Calculation.band_structure_phonon)
            for n_segment in range(len(kpoints)):
                sec_segment = sec_bandstructure.m_create(BandEnergies)
                sec_segment.kpoints = kpoints[n_segment]
                sec_segment.energies = np.reshape(
                    bandstructure[n_segment], (1, *np.shape(bandstructure[n_segment]))) * ureg.millielectron_volt

        self.apl_parser.mainfile = self.get_aflow_file('aflow.apl.thermodynamic_properties.out')

        sec_workflow = self.archive.m_create(Workflow)
        sec_workflow.type = 'phonon'
        sec_workflow.run_ref = sec_run

        sec_phonon = sec_workflow.m_create(Phonon)
        sec_phonon.force_calculator = 'vasp'
        mesh = self.aflowin_parser.get('aflow_apl_dosmesh')
        if mesh is not None:
            try:
                cell = Cell.fromcellpar(self.aflowin_parser.geometry)
                sec_phonon.mesh_density = np.product([int(m) for m in mesh.split('x')]) / cell.volume
            except Exception:
                pass

        self.apl_parser.mainfile = self.get_aflow_file('aflow.apl.group_velocities.out')
        group_velocity = self.apl_parser.get('apl_group_velocity')
        if group_velocity is not None:
            try:
                qpoints = self.apl_parser.apl_qpoints
                qpoints = np.reshape(qpoints, (len(qpoints) // 4, 4))
                sec_phonon.qpoints = np.transpose(np.transpose(qpoints)[1:])
                group_velocity = np.reshape(
                    group_velocity, (len(qpoints), len(group_velocity) // len(qpoints)))
                group_velocity = np.transpose(np.transpose(group_velocity)[1:])
                sec_phonon.group_velocity = np.reshape(group_velocity, (
                    len(group_velocity), len(group_velocity[0]) // 3, 3)) * ureg.kilometer / ureg.second
            except Exception:
                pass

        self.apl_parser.mainfile = self.get_aflow_file('aflow.apl.thermodynamic_properties.out.xz')
        apl_thermo = self.apl_parser.get('apl_thermo')
        if apl_thermo is not None:
            apl_thermo = np.transpose(np.reshape(apl_thermo, (len(apl_thermo) // 6, 6)))
            sec_thermo = sec_workflow.m_create(WorkflowThermodynamics)
            sec_thermo.temperature = apl_thermo[0] * ureg.kelvin
            sec_thermo.internal_energy = apl_thermo[2] * ureg.millielectron_volt
            sec_thermo.helmholtz_free_energy = apl_thermo[3] * ureg.millielectron_volt
            sec_thermo.entropy = apl_thermo[4] * ureg.boltzmann_constant
            sec_thermo.heat_capacity_c_v = apl_thermo[5] * ureg.boltzmann_constant

        # TODO parse systems for each displacements

        # TODO parse displacements, force constants, dynamical matrix

    def parse(self, filepath, archive, logger):
        self.filepath = os.path.abspath(filepath)
        self.archive = archive
        self.maindir = os.path.dirname(self.filepath)
        self.logger = logger if logger is not None else logging

        self.init_parser()

        sec_run = self.archive.m_create(Run)
        sec_run.program = Program(
            name='AFlow', version=self.aflow_data.get('aflow_version', 'unknown'))

        # parse run metadata
        run_quantities = ['aurl', 'auid', 'data_api', 'data_source', 'loop']
        for key in run_quantities:
            val = self.aflow_data.get(key)
            if val is not None:
                setattr(sec_run, 'x_aflow_%s' % key, val)

        # TODO The OUTCAR file will be read by the vasp parser and so the complete
        # metadata for both system and method should be filled in by vasp parser.
        # parse structure from aflow_data
        sec_system = sec_run.m_create(System)
        sec_system.atoms = Atoms()
        lattice_parameters = self.aflow_data.get('geometry')
        if lattice_parameters is not None:
            cell = self.aflow_data.get('cell', Cell.fromcellpar(lattice_parameters))
            sec_system.atoms.lattice_vectors = cell.array * ureg.angstrom
            sec_system.atoms.periodic = [True, True, True]
        species = self.aflow_data.get('species', [])
        atom_labels = []
        for n, specie in enumerate(species):
            atom_labels += [specie] * self.aflow_data['composition'][n]
        sec_system.atoms.labels = atom_labels
        sec_system.atoms.positions = self.aflow_data.get('positions_cartesian', []) * ureg.angstrom

        # parse system metadata from aflow_data
        system_quantities = [
            'compound', 'prototype', 'nspecies', 'natoms', 'natoms_orig', 'composition',
            'density', 'density_orig', 'scintillation_attenuation_length', 'stoichiometry',
            'species', 'geometry', 'geometry_orig', 'volume_cell', 'volume_atom',
            'volume_cell_orig', 'volume_atom_orig', 'n_sg', 'sg', 'sg2', 'spacegroup_orig',
            'spacegroup_relax', 'Bravais_lattice_orig', 'lattice_variation_orig',
            'lattice_system_orig', 'Pearson_symbol_orig', 'Bravais_lattice_relax',
            'lattice_variation_relax', 'lattice_system_relax', 'Pearson_symbol_relax',
            'crystal_family_orig', 'crystal_system_orig', 'crystal_class_orig',
            'point_group_Hermann_Mauguin_orig', 'point_group_Schoenflies_orig',
            'point_group_orbifold_orig', 'point_group_type_orig', 'point_group_order_orig',
            'point_group_structure_orig', 'Bravais_lattice_lattice_type_orig',
            'Bravais_lattice_lattice_variation_type_orig', 'Bravais_lattice_lattice_system_orig',
            'Bravais_superlattice_lattice_type_orig', 'Bravais_superlattice_lattice_variation_type_orig',
            'Bravais_superlattice_lattice_system_orig', 'Pearson_symbol_superlattice_orig',
            'reciprocal_geometry_orig', 'reciprocal_volume_cell_orig', 'reciprocal_lattice_type_orig',
            'reciprocal_lattice_variation_type_orig', 'Wyckoff_letters_orig',
            'Wyckoff_multiplicities_orig', 'Wyckoff_site_symmetries_orig',
            'crystal_family', 'crystal_system', 'crystal_class', 'point_group_Hermann_Mauguin',
            'point_group_Schoenflies', 'point_group_orbifold', 'point_group_type', 'point_group_order',
            'point_group_structure', 'Bravais_lattice_lattice_type', 'Bravais_lattice_lattice_variation_type',
            'Bravais_lattice_lattice_system', 'Bravais_superlattice_lattice_type',
            'Bravais_superlattice_lattice_variation_type', 'Bravais_superlattice_lattice_system',
            'Pearson_symbol_superlattice', 'reciprocal_geometry', 'reciprocal_volume_cell',
            'reciprocal_lattice_type', 'reciprocal_lattice_variation_type', 'Wyckoff_letters',
            'Wyckoff_multiplicities', 'Wyckoff_site_symmetries', 'prototype_label_orig',
            'prototype_params_list_orig', 'prototype_params_values_orig', 'prototype_label_relax',
            'prototype_params_list_relax', 'prototype_params_values_relax']
        for key in system_quantities:
            val = self.aflow_data.get(key)
            if val is not None:
                if 'Wyckoff' in key:
                    val = np.array(val, dtype=object)
                setattr(sec_system, 'x_aflow_%s' % key, val)

        # parse method metadata from self.aflow_data
        method_quantities = [
            'code', 'species_pp', 'n_dft_type', 'dft_type', 'dft_type', 'species_pp_version',
            'species_pp_ZVAL', 'species_pp_AUID', 'ldau_type', 'ldau_l', 'ldau_u', 'ldau_j',
            'valence_cell_iupac', 'valence_cell_std', 'energy_cutoff',
            'delta_electronic_energy_convergence', 'delta_electronic_energy_threshold',
            'kpoints_relax', 'kpoints_static', 'n_kpoints_bands_path', 'kpoints_bands_path',
            'kpoints_bands_nkpts']
        sec_method = sec_run.m_create(Method)
        for key in method_quantities:
            val = self.aflow_data.get(key)
            if val is not None:
                setattr(sec_method, 'x_aflow_%s' % key, val)

        # parse basic calculation quantities from self.aflow_data
        sec_scc = sec_run.m_create(Calculation)
        sec_scc.energy = Energy()
        sec_scc.forces = Forces()
        sec_thermo = sec_scc.m_create(Thermodynamics)
        if self.aflow_data.get('energy_cell') is not None:
            sec_scc.energy.total = EnergyEntry(value=self.aflow_data['energy_cell'] * ureg.eV)
        if self.aflow_data.get('forces') is not None:
            sec_scc.forces.total = ForcesEntry(value=self.aflow_data['forces'] * ureg.eV / ureg.angstrom)
        if self.aflow_data.get('enthalpy_cell') is not None:
            sec_thermo.enthalpy = self.aflow_data['enthalpy_cell'] * ureg.eV
        if self.aflow_data.get('entropy_cell') is not None:
            sec_thermo.entropy = self.aflow_data['entropy_cell'] * ureg.eV / ureg.K
        if self.aflow_data.get('calculation_time') is not None:
            sec_scc.time_calculation = self.aflow_data['calculation_time'] * ureg.s
        calculation_quantities = [
            'stress_tensor', 'pressure_residual', 'Pulay_stress', 'Egap', 'Egap_fit', 'Egap_type',
            'enthalpy_formation_cell', 'entropic_temperature', 'PV', 'spin_cell', 'spinD',
            'spinF', 'calculation_memory', 'calculation_cores', 'nbondxx',
            'agl_thermal_conductivity_300K', 'agl_debye', 'agl_acoustic_debye', 'agl_gruneisen',
            'agl_heat_capacity_Cv_300K', 'agl_heat_capacity_Cp_300K', 'agl_thermal_expansion_300K',
            'agl_bulk_modulus_static_300K', 'agl_bulk_modulus_isothermal_300K', 'agl_poisson_ratio_source',
            'agl_vibrational_free_energy_300K_cell', 'agl_vibrational_free_energy_300K_atom',
            'agl_vibrational_entropy_300K_cell', 'agl_vibrational_entropy_300K_atom',
            'ael_poisson_ratio', 'ael_bulk_modulus_voigt', 'ael_bulk_modulus_reuss',
            'ael_shear_modulus_voigt', 'ael_shear_modulus_reuss', 'ael_bulk_modulus_vrh',
            'ael_shear_modulus_vrh', 'ael_elastic_anisotropy', 'ael_youngs_modulus_vrh',
            'ael_speed_sound_transverse', 'ael_speed_sound_longitudinal', 'ael_speed_sound_average',
            'ael_pughs_modulus_ratio', 'ael_debye_temperature', 'ael_applied_pressure',
            'ael_average_external_pressure', 'ael_stiffness_tensor', 'ael_compliance_tensor',
            'bader_net_charges', 'bader_atomic_volumes', 'n_files', 'files', 'node_CPU_Model',
            'node_CPU_Cores', 'node_CPU_MHz', 'node_RAM_GB', 'catalog', 'aflowlib_version',
            'aflowlib_date']
        for key in calculation_quantities:
            val = self.aflow_data.get(key)
            if val is not None:
                setattr(sec_scc, 'x_aflow_%s' % key, val)

        for module in self.aflow_data.get('loop', []):
            if module == 'ael':
                self.parse_ael()
            elif module == 'agl':
                self.parse_agl()
            elif module == 'apl':
                self.parse_apl()



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD.
# See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import os
import json
import numpy as np
from typing import List
import datetime

try:
    import asr
    from asr.core.cache import get_cache  # pylint: disable=E0611,E0401
    from asr.core.record import Record  # pylint: disable=E0611,E0401
except Exception:
    pass

from nomad.units import ureg
from nomad.datamodel import EntryArchive
from nomad.datamodel.metainfo.simulation.run import Run, Program, TimeRun
from nomad.datamodel.metainfo.simulation.system import System, Atoms
from nomad.datamodel.metainfo.simulation.calculation import (
    Calculation, BandStructure, BandEnergies, Energy, EnergyEntry, Forces, ForcesEntry,
    Stress, StressEntry)
from nomad.datamodel.metainfo.workflow import Workflow
from .metainfo.asr import (
    x_asr_resources, x_asr_metadata,
    x_asr_run_specification, x_asr_parameters, x_asr_code, x_asr_codes, x_asr_dependencies,
    x_asr_dependency)


class ASRRecord:
    def __init__(self, record=None):
        self._record = record
        self._converted = False

    @property
    def archive(self):
        if self._archive is None:
            self._archive = EntryArchive()
        if not self._converted:
            self.to_archive()
        return self._archive

    @property
    def record(self):
        return self._record

    @record.setter
    def record(self, value):
        self._record = value
        self._archive = None
        self._converted = False

    def _parse_system(self, atoms):
        system = self._archive.run[-1].m_create(System)
        system.atoms = Atoms(
            positions=atoms.get_positions() * ureg.angstrom,
            lattice_vectors=atoms.get_cell().array * ureg.angstrom,
            periodic=[True, True, True],
            labels=atoms.get_chemical_symbols())

    def _parse_c2db_relax(self):
        result = self.record.result
        for image in result.images:
            self._parse_system(image)
        workflow = self._archive.m_create(Workflow)
        workflow.type = 'geometry_optimization'
        workflow.run_ref = self._archive.run[-1]

        calc = self._archive.run[-1].m_create(Calculation)
        calc.system_ref = self._archive.run[-1].system[-1]
        calc.energy = Energy(total=EnergyEntry(value=result.etot * ureg.eV))
        calc.forces = Forces(total=ForcesEntry(value=result.forces * ureg.eV / ureg.angstrom))
        stress = np.zeros((3, 3))
        stress[0][0] = result.stress[0]
        stress[1][1] = result.stress[1]
        stress[2][2] = result.stress[2]
        stress[0][1] = stress[1][0] = result.stress[3]
        stress[0][2] = stress[2][0] = result.stress[4]
        stress[1][2] = stress[2][1] = result.stress[5]
        calc.stress = Stress(total=StressEntry(value=stress))

    def _parse_c2db_phonopy(self):
        workflow = self._archive.m_create(Workflow)
        workflow.type = 'phonon'
        workflow.run_ref = self._archive.run[-1]

        bands = self.record.result.data.get('omega_kl')

        path = self.record.result.data.get('path')
        hisym_kpts = [list(p) for p in path.special_points.values()]
        labels = list(path.special_points.keys())
        endpoints = []
        calc = self._archive.run[-1].m_create(Calculation)
        bandstructure = calc.m_create(BandStructure, Calculation.band_structure_phonon)
        for i, qpoint in enumerate(path.kpts):
            if list(qpoint) in hisym_kpts:
                endpoints.append(i)
            if len(endpoints) < 2:
                continue
            sec_segment = bandstructure.m_create(BandEnergies)
            energies = bands[endpoints[0]: endpoints[1] + 1]
            sec_segment.energies = np.reshape(energies, (1, *np.shape(energies)))
            sec_segment.kpoints = path.kpts[endpoints[0]: endpoints[1] + 1]
            sec_segment.endpoints_labels = [labels[hisym_kpts.index(list(path.kpts[i]))] for i in endpoints]
            endpoints = [i]

    def _parse_run(self):
        run = self._archive.m_create(Run)
        run.program = Program(name='ASR', version=asr.__version__)

        if self.record.resources is not None:
            run.time_run = TimeRun(
                date_start=self.record.resources.execution_start,
                date_end=self.record.resources.execution_end)
        resources = run.m_create(x_asr_resources)
        for key, val in self.record.resources.__dict__.items():
            try:
                setattr(resources, 'x_asr_%s' % key, val)
            except Exception:
                pass

        if self.record.metadata is not None:
            metadata = run.m_create(x_asr_metadata)
            metadata.x_asr_created = (
                self.record.metadata.created - datetime.datetime(1970, 1, 1)).total_seconds()
            metadata.x_asr_modified = (
                self.record.metadata.modified - datetime.datetime(1970, 1, 1)).total_seconds()
            metadata.x_asr_directory = self.record.metadata.directory

        # misc
        for key, val in self.record.__dict__.items():
            if hasattr(val, '__dict__'):
                continue
            try:
                setattr(run, 'x_asr_%s' % key, val)
            except Exception:
                pass

        if self.record.run_specification is not None:
            # parse original system info
            atoms = self.record.run_specification.parameters.atoms
            self._parse_system(atoms)
            run_spec = run.m_create(x_asr_run_specification)
            for key, val in self.record.run_specification.__dict__.items():
                if hasattr(val, '__dict__'):
                    continue
                try:
                    setattr(run_spec, 'x_asr_%s' % key, val)
                except Exception:
                    pass
            parameters = run_spec.m_create(x_asr_parameters)
            for key, val in self.record.run_specification.parameters.__dict__.items():
                if hasattr(val, '__dict__'):
                    continue
                try:
                    setattr(parameters, 'x_asr_%s' % key, val)
                except Exception:
                    pass
            codes = run_spec.m_create(x_asr_codes)
            for entry in self.record.run_specification.codes.codes:
                code = codes.m_create(x_asr_code)
                code.x_asr_package = entry.package
                code.x_asr_version = entry.version
                code.x_asr_git_hash = entry.git_hash

        if self.record.dependencies is not None:
            dependencies = run.m_create(x_asr_dependencies)
            for dep in self.record.dependencies.deps:
                dependency = dependencies.m_create(x_asr_dependency)
                dependency.x_asr_uid = dep.uid
                dependency.x_asr_revision = dep.revision

    def to_archive(self):
        self._parse_run()
        if self.record.name == 'asr.c2db.relax':
            self._parse_c2db_relax()
        elif self.record.name == 'asr.c2db.phonopy':
            self._parse_c2db_phonopy()
        self._converted = True
        return self._archive


class ASRParser:
    def __init__(self):
        pass

    def parse(self, mainfile: str, archive: EntryArchive, logger=None):
        with open(mainfile, 'rt') as f:
            archive_data = json.load(f)
            archive.m_update_from_dict(archive_data)


def asr_to_archives(directory: str, recipes: List[str] = None):
    '''
    Converts the asr results for the specified recipes under the given directory to the
    nomad archive format.
    '''
    # record can only be fetched on the directory
    try:
        cwd = os.getcwd()
        os.chdir(directory)
        cache = get_cache()

        records: List[Record] = []
        if recipes is None:
            records = cache.select()
        else:
            for recipe in recipes:
                records.extend(cache.select(name=recipe))
        asr_record = ASRRecord()
        for record in records:
            asr_record.record = record
            with open('archive_%s.json' % record.uid, 'w') as f:
                json.dump(asr_record.archive.m_to_dict(), f, indent=4)
    except Exception:
        pass
    finally:
        os.chdir(cwd)



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD.
# See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import numpy as np            # pylint: disable=unused-import
import typing                 # pylint: disable=unused-import
from nomad.metainfo import (  # pylint: disable=unused-import
    MSection, MCategory, Category, Package, Quantity, Section, SubSection, SectionProxy,
    Reference
)
from nomad.datamodel.metainfo import simulation
from nomad.datamodel.metainfo import workflow


m_package = Package()


class x_namd_mdin_input_output_files(MCategory):
    '''
    Parameters of mdin belonging to x_namd_section_control_parameters.
    '''

    m_def = Category()


class x_namd_mdin_control_parameters(MCategory):
    '''
    Parameters of mdin belonging to x_namd_section_control_parameters.
    '''

    m_def = Category()


class x_namd_mdin_method(MCategory):
    '''
    Parameters of mdin belonging to section method.
    '''

    m_def = Category()


class x_namd_mdout_single_configuration_calculation(MCategory):
    '''
    Parameters of mdout belonging to section_single_configuration_calculation.
    '''

    m_def = Category()


class x_namd_mdout_method(MCategory):
    '''
    Parameters of mdin belonging to section method.
    '''

    m_def = Category()


class x_namd_mdout_run(MCategory):
    '''
    Parameters of mdin belonging to settings run.
    '''

    m_def = Category()


class x_namd_mdin_run(MCategory):
    '''
    Parameters of mdin belonging to settings run.
    '''

    m_def = Category()


class x_namd_section_input_output_files(MSection):
    '''
    Section to store input and output file names
    '''

    m_def = Section(validate=False)

    x_namd_inout_file_structure = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD input topology file.
        ''')

    x_namd_inout_file_traj_coord = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD output trajectory file.
        ''')

    x_namd_inout_file_traj_vel = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD output file for velocities in the trajectory.
        ''')

    x_namd_inout_file_traj_force = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD output file for forces in the trajectory.
        ''')

    x_namd_inout_file_output_coord = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD output coordinates file.
        ''')

    x_namd_inout_file_output_vel = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD output velocities file.
        ''')

    x_namd_inout_file_output_force = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD output forces file.
        ''')

    x_namd_inout_file_input_coord = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD input coordinates file.
        ''')

    x_namd_inout_file_input_vel = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD input velocities file.
        ''')

    x_namd_inout_file_restart_coord = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD restart coordinates file.
        ''')

    x_namd_inout_file_restart_vel = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD restart velocities file.
        ''')

    x_namd_inout_file_fftw_datafile = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD FFTW data file.
        ''')

    x_namd_inout_file_mdlog = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD MD output log file.
        ''')


class x_namd_section_control_parameters(MSection):
    '''
    Section to store the input and output control parameters
    '''

    m_def = Section(validate=False)

    x_namd_inout_control_timestep = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD running environment and control parameters.
        ''')

    x_namd_inout_control_number_of_steps = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD running environment and control parameters.
        ''')

    x_namd_inout_control_steps_per_cycle = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD running environment and control parameters.
        ''')

    x_namd_inout_control_periodic_cell_basis_1 = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD running environment and control parameters.
        ''')

    x_namd_inout_control_periodic_cell_basis_2 = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD running environment and control parameters.
        ''')

    x_namd_inout_control_periodic_cell_basis_3 = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD running environment and control parameters.
        ''')

    x_namd_inout_control_periodic_cell_center = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD running environment and control parameters.
        ''')

    x_namd_inout_control_load_balancer = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD running environment and control parameters.
        ''')

    x_namd_inout_control_load_balancing_strategy = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD running environment and control parameters.
        ''')

    x_namd_inout_control_ldb_period = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD running environment and control parameters.
        ''')

    x_namd_inout_control_first_ldb_timestep = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD running environment and control parameters.
        ''')

    x_namd_inout_control_last_ldb_timestep = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD running environment and control parameters.
        ''')

    x_namd_inout_control_ldb_background_scaling = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD running environment and control parameters.
        ''')

    x_namd_inout_control_hom_background_scaling = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD running environment and control parameters.
        ''')

    x_namd_inout_control_pme_background_scaling = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD running environment and control parameters.
        ''')

    x_namd_inout_control_min_atoms_per_patch = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD running environment and control parameters.
        ''')

    x_namd_inout_control_initial_temperature = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD running environment and control parameters.
        ''')

    x_namd_inout_control_center_of_mass_moving_initially = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD running environment and control parameters.
        ''')

    x_namd_inout_control_dielectric = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD running environment and control parameters.
        ''')

    x_namd_inout_control_excluded_species_or_groups = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD running environment and control parameters.
        ''')

    x_namd_inout_control_1_4_electrostatics_scale = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD running environment and control parameters.
        ''')

    x_namd_inout_control_traj_dcd_filename = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD running environment and control parameters.
        ''')

    x_namd_inout_control_traj_dcd_frequency = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD running environment and control parameters.
        ''')

    x_namd_inout_control_traj_dcd_first_step = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD running environment and control parameters.
        ''')

    x_namd_inout_control_velocity_dcd_filename = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD running environment and control parameters.
        ''')

    x_namd_inout_control_velocity_dcd_frequency = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD running environment and control parameters.
        ''')

    x_namd_inout_control_velocity_dcd_first_step = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD running environment and control parameters.
        ''')

    x_namd_inout_control_force_dcd_filename = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD running environment and control parameters.
        ''')

    x_namd_inout_control_force_dcd_frequency = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD running environment and control parameters.
        ''')

    x_namd_inout_control_force_dcd_first_step = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD running environment and control parameters.
        ''')

    x_namd_inout_control_output_filename = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD running environment and control parameters.
        ''')

    x_namd_inout_control_binary_output = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD running environment and control parameters.
        ''')

    x_namd_inout_control_restart_filename = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD running environment and control parameters.
        ''')

    x_namd_inout_control_restart_frequency = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD running environment and control parameters.
        ''')

    x_namd_inout_control_binary_restart = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD running environment and control parameters.
        ''')

    x_namd_inout_control_switching = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD running environment and control parameters.
        ''')

    x_namd_inout_control_switching_on = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD running environment and control parameters.
        ''')

    x_namd_inout_control_switching_off = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD running environment and control parameters.
        ''')

    x_namd_inout_control_pairlist_distance = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD running environment and control parameters.
        ''')

    x_namd_inout_control_pairlist_shrink_rate = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD running environment and control parameters.
        ''')

    x_namd_inout_control_pairlist_grow_rate = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD running environment and control parameters.
        ''')

    x_namd_inout_control_pairlist_trigger = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD running environment and control parameters.
        ''')

    x_namd_inout_control_pairlists_per_cycle = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD running environment and control parameters.
        ''')

    x_namd_inout_control_pairlists = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD running environment and control parameters.
        ''')

    x_namd_inout_control_margin = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD running environment and control parameters.
        ''')

    x_namd_inout_control_hydrogen_group_cutoff = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD running environment and control parameters.
        ''')

    x_namd_inout_control_patch_dimension = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD running environment and control parameters.
        ''')

    x_namd_inout_control_energy_output_steps = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD running environment and control parameters.
        ''')

    x_namd_inout_control_crossterm_energy = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD running environment and control parameters.
        ''')

    x_namd_inout_control_timing_output_steps = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD running environment and control parameters.
        ''')

    x_namd_inout_control_velocity_rescale_freq = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD running environment and control parameters.
        ''')

    x_namd_inout_control_velocity_rescale_temp = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD running environment and control parameters.
        ''')

    x_namd_inout_control_velocity_reassignment_freq = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD running environment and control parameters.
        ''')

    x_namd_inout_control_velocity_reassignment_temp = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD running environment and control parameters.
        ''')

    x_namd_inout_control_velocity_reassignment_incr = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD running environment and control parameters.
        ''')

    x_namd_inout_control_velocity_reassignment_hold = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD running environment and control parameters.
        ''')

    x_namd_inout_control_lowe_andersen_dynamics = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD running environment and control parameters.
        ''')

    x_namd_inout_control_lowe_andersen_temperature = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD running environment and control parameters.
        ''')

    x_namd_inout_control_lowe_andersen_rate = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD running environment and control parameters.
        ''')

    x_namd_inout_control_lowe_andersen_cutoff = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD running environment and control parameters.
        ''')

    x_namd_inout_control_langevin_dynamics = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD running environment and control parameters.
        ''')

    x_namd_inout_control_langevin_temperature = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD running environment and control parameters.
        ''')

    x_namd_inout_control_langevin_integrator = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD running environment and control parameters.
        ''')

    x_namd_inout_control_langevin_damping_file = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD running environment and control parameters.
        ''')

    x_namd_inout_control_langevin_damping_column = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD running environment and control parameters.
        ''')

    x_namd_inout_control_langevin_damping_coefficient_unit = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD running environment and control parameters.
        ''')

    x_namd_inout_control_langevin_dynamics_not_applied_to = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD running environment and control parameters.
        ''')

    x_namd_inout_control_temperature_coupling = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD running environment and control parameters.
        ''')

    x_namd_inout_control_coupling_temperature = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD running environment and control parameters.
        ''')

    x_namd_inout_control_berendsen_pressure_coupling = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD running environment and control parameters.
        ''')

    x_namd_inout_control_berendsen_compressibility_estimate = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD running environment and control parameters.
        ''')

    x_namd_inout_control_berendsen_relaxation_time = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD running environment and control parameters.
        ''')

    x_namd_inout_control_berendsen_coupling_frequency = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD running environment and control parameters.
        ''')

    x_namd_inout_control_langevin_piston_pressure_control = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD running environment and control parameters.
        ''')

    x_namd_inout_control_target_pressure = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD running environment and control parameters.
        ''')

    x_namd_inout_control_langevin_oscillation_period = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD running environment and control parameters.
        ''')

    x_namd_inout_control_langevin_decay_time = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD running environment and control parameters.
        ''')

    x_namd_inout_control_langevin_piston_temperature = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD running environment and control parameters.
        ''')

    x_namd_inout_control_pressure_control = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD running environment and control parameters.
        ''')

    x_namd_inout_control_initial_strain_rate = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD running environment and control parameters.
        ''')

    x_namd_inout_control_cell_fluctuation = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD running environment and control parameters.
        ''')

    x_namd_inout_control_particle_mesh_ewald = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD running environment and control parameters.
        ''')

    x_namd_inout_control_pme_tolerance = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD running environment and control parameters.
        ''')

    x_namd_inout_control_pme_ewald_coefficient = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD running environment and control parameters.
        ''')

    x_namd_inout_control_pme_interpolation_order = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD running environment and control parameters.
        ''')

    x_namd_inout_control_pme_grid_dimensions = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD running environment and control parameters.
        ''')

    x_namd_inout_control_pme_maximum_grid_spacing = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD running environment and control parameters.
        ''')

    x_namd_inout_control_fftw_data_file = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD running environment and control parameters.
        ''')

    x_namd_inout_control_full_electrostatic_evaluation_frequency = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD running environment and control parameters.
        ''')

    x_namd_inout_control_minimization = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD running environment and control parameters.
        ''')

    x_namd_inout_control_velocity_quenching = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD running environment and control parameters.
        ''')

    x_namd_inout_control_verlet_integrator = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD running environment and control parameters.
        ''')

    x_namd_inout_control_random_number_seed = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD running environment and control parameters.
        ''')

    x_namd_inout_control_use_hydrogen_bonds = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD running environment and control parameters.
        ''')

    x_namd_inout_control_coordinate_pdb = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD running environment and control parameters.
        ''')

    x_namd_inout_control_structure_file = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD running environment and control parameters.
        ''')

    x_namd_inout_control_parameter_file = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD running environment and control parameters.
        ''')

    x_namd_inout_control_number_of_parameters = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        NAMD running environment and control parameters.
        ''')

    x_namd_inout_control_parameters = Quantity(
        type=str,
        shape=['x_namd_inout_control_number_of_parameters'],
        description='''
        NAMD running environment and control parameters.
        ''')


class x_namd_section_atom_to_atom_type_ref(MSection):
    '''
    Section to store atom label to atom type definition list
    '''

    m_def = Section(validate=False)

    x_namd_atom_to_atom_type_ref = Quantity(
        type=np.dtype(np.int64),
        shape=['number_of_atoms_per_type'],
        description='''
        Reference to the atoms of each atom type.
        ''')


class x_namd_section_single_configuration_calculation(MSection):
    '''
    section for gathering values for MD steps
    '''

    m_def = Section(validate=False)


class System(simulation.system.System):

    m_def = Section(validate=False, extends_base_section=True)

    x_namd_atom_positions_image_index = Quantity(
        type=np.dtype(np.int32),
        shape=['number_of_atoms', 3],
        unit='dimensionless',
        description='''
        PBC image flag index.
        ''')

    x_namd_atom_positions_scaled = Quantity(
        type=np.dtype(np.float64),
        shape=['number_of_atoms', 3],
        unit='dimensionless',
        description='''
        Position of the atoms in a scaled format [0, 1].
        ''')

    x_namd_atom_positions_wrapped = Quantity(
        type=np.dtype(np.float64),
        shape=['number_of_atoms', 3],
        unit='meter',
        description='''
        Position of the atoms wrapped back to the periodic box.
        ''')

    x_namd_lattice_lengths = Quantity(
        type=np.dtype(np.float64),
        shape=[3],
        description='''
        Lattice dimensions in a vector. Vector includes [a, b, c] lengths.
        ''')

    x_namd_lattice_angles = Quantity(
        type=np.dtype(np.float64),
        shape=[3],
        description='''
        Angles of lattice vectors. Vector includes [alpha, beta, gamma] in degrees.
        ''')

    x_namd_dummy = Quantity(
        type=str,
        shape=[],
        description='''
        dummy
        ''')

    x_namd_mdin_finline = Quantity(
        type=str,
        shape=[],
        description='''
        finline in mdin
        ''')

    x_namd_traj_timestep_store = Quantity(
        type=str,
        shape=[],
        description='''
        tmp
        ''')

    x_namd_traj_number_of_atoms_store = Quantity(
        type=str,
        shape=[],
        description='''
        tmp
        ''')

    x_namd_traj_box_bound_store = Quantity(
        type=str,
        shape=[],
        description='''
        tmp
        ''')

    x_namd_traj_box_bounds_store = Quantity(
        type=str,
        shape=[],
        description='''
        tmp
        ''')

    x_namd_traj_variables_store = Quantity(
        type=str,
        shape=[],
        description='''
        tmp
        ''')

    x_namd_traj_atoms_store = Quantity(
        type=str,
        shape=[],
        description='''
        tmp
        ''')


class MolecularDynamics(workflow.MolecularDynamics):

    m_def = Section(validate=False, extends_base_section=True)

    x_namd_barostat_target_pressure = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='pascal',
        description='''
        MD barostat target pressure.
        ''')

    x_namd_barostat_tau = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='second',
        description='''
        MD barostat relaxation time.
        ''')

    x_namd_barostat_type = Quantity(
        type=str,
        shape=[],
        description='''
        MD barostat type, valid values are defined in the barostat_type wiki page.
        ''')

    x_namd_integrator_dt = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='second',
        description='''
        MD integration time step.
        ''')

    x_namd_integrator_type = Quantity(
        type=str,
        shape=[],
        description='''
        MD integrator type, valid values are defined in the integrator_type wiki page.
        ''')

    x_namd_periodicity_type = Quantity(
        type=str,
        shape=[],
        description='''
        Periodic boundary condition type in the sampling (non-PBC or PBC).
        ''')

    x_namd_langevin_gamma = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='second',
        description='''
        Langevin thermostat damping factor.
        ''')

    x_namd_number_of_steps_requested = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Number of requested MD integration time steps.
        ''')

    x_namd_thermostat_level = Quantity(
        type=str,
        shape=[],
        description='''
        MD thermostat level (see wiki: single, multiple, regional).
        ''')

    x_namd_thermostat_target_temperature = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='kelvin',
        description='''
        MD thermostat target temperature.
        ''')

    x_namd_thermostat_tau = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='second',
        description='''
        MD thermostat relaxation time.
        ''')

    x_namd_thermostat_type = Quantity(
        type=str,
        shape=[],
        description='''
        MD thermostat type, valid values are defined in the thermostat_type wiki page.
        ''')


class AtomParameters(simulation.method.AtomParameters):

    m_def = Section(validate=False, extends_base_section=True)

    x_namd_atom_name = Quantity(
        type=str,
        shape=[],
        description='''
        Atom name of an atom in topology definition.
        ''')

    x_namd_atom_type = Quantity(
        type=str,
        shape=[],
        description='''
        Atom type of an atom in topology definition.
        ''')

    x_namd_atom_element = Quantity(
        type=str,
        shape=[],
        description='''
        Atom type of an atom in topology definition.
        ''')

    x_namd_atom_type_element = Quantity(
        type=str,
        shape=[],
        description='''
        Element symbol of an atom type.
        ''')

    x_namd_atom_type_radius = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        van der Waals radius of an atom type.
        ''')

    number_of_atoms_per_type = Quantity(
        type=int,
        shape=[],
        description='''
        Number of atoms involved in this type.
        ''')


class Interaction(simulation.method.Interaction):

    m_def = Section(validate=False, extends_base_section=True)

    x_namd_interaction_atom_to_atom_type_ref = Quantity(
        type=simulation.method.AtomParameters,
        shape=['number_of_atoms_per_interaction'],
        description='''
        Reference to the atom type of each interaction atoms.
        ''')

    x_namd_number_of_defined_pair_interactions = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        Number of defined pair interactions (L-J pairs).
        ''')

    x_namd_pair_interaction_atom_type_ref = Quantity(
        type=simulation.method.AtomParameters,
        shape=['x_namd_number_of_defined_pair_interactions', 'number_of_atoms_per_interaction'],
        description='''
        Reference to the atom type for pair interactions.
        ''')

    x_namd_pair_interaction_parameters = Quantity(
        type=np.dtype(np.float64),
        shape=['x_namd_number_of_defined_pair_interactions', 2],
        description='''
        Pair interactions parameters.
        ''')

    x_namd_molecule_interaction_atom_to_atom_type_ref = Quantity(
        type=simulation.method.AtomParameters,
        shape=['number_of_atoms_per_interaction'],
        description='''
        Reference to the atom type of each molecule interaction atoms.
        ''')

    x_namd_number_of_defined_molecule_pair_interactions = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        Number of defined pair interactions within a molecule (L-J pairs).
        ''')

    x_namd_pair_molecule_interaction_parameters = Quantity(
        type=np.dtype(np.float64),
        shape=['number_of_defined_molecule_pair_interactions', 2],
        description='''
        Molecule pair interactions parameters.
        ''')

    x_namd_pair_molecule_interaction_to_atom_type_ref = Quantity(
        type=simulation.method.AtomParameters,
        shape=['x_namd_number_of_defined_pair_interactions', 'number_of_atoms_per_interaction'],
        description='''
        Reference to the atom type for pair interactions within a molecule.
        ''')


class Run(simulation.run.Run):

    m_def = Section(validate=False, extends_base_section=True)

    x_namd_program_version_date = Quantity(
        type=str,
        shape=[],
        description='''
        Program version date.
        ''')

    x_namd_parallel_task_nr = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Program task no.
        ''')

    x_namd_build_osarch = Quantity(
        type=str,
        shape=[],
        description='''
        Program Build OS/ARCH
        ''')

    x_namd_program_build_date = Quantity(
        type=str,
        shape=[],
        description='''
        Program Build date
        ''')

    x_namd_program_citation = Quantity(
        type=str,
        shape=[],
        description='''
        Program citations
        ''')

    x_namd_number_of_tasks = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Number of tasks in parallel program (MPI).
        ''')

    x_namd_program_module_version = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD program module version.
        ''')

    x_namd_program_license = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD program license.
        ''')

    x_namd_xlo_xhi = Quantity(
        type=str,
        shape=[],
        description='''
        test
        ''')

    x_namd_data_file_store = Quantity(
        type=str,
        shape=[],
        description='''
        Filename of data file
        ''')

    x_namd_program_working_path = Quantity(
        type=str,
        shape=[],
        description='''
        tmp
        ''')

    x_namd_program_execution_host = Quantity(
        type=str,
        shape=[],
        description='''
        tmp
        ''')

    x_namd_program_execution_path = Quantity(
        type=str,
        shape=[],
        description='''
        tmp
        ''')

    x_namd_program_module = Quantity(
        type=str,
        shape=[],
        description='''
        tmp
        ''')

    x_namd_program_execution_date = Quantity(
        type=str,
        shape=[],
        description='''
        tmp
        ''')

    x_namd_program_execution_time = Quantity(
        type=str,
        shape=[],
        description='''
        tmp
        ''')

    x_namd_mdin_header = Quantity(
        type=str,
        shape=[],
        description='''
        tmp
        ''')

    x_namd_mdin_wt = Quantity(
        type=str,
        shape=[],
        description='''
        tmp
        ''')

    x_namd_section_input_output_files = SubSection(
        sub_section=SectionProxy('x_namd_section_input_output_files'),
        repeats=True)

    x_namd_section_control_parameters = SubSection(
        sub_section=SectionProxy('x_namd_section_control_parameters'),
        repeats=True)


class Calculation(simulation.calculation.Calculation):

    m_def = Section(validate=False, extends_base_section=True)

    x_namd_section_single_configuration_calculation = SubSection(
        sub_section=SectionProxy('x_namd_section_single_configuration_calculation'),
        repeats=True)



#!/usr/bin/env python3
"""Main file of yaml2nxdl tool.
Users create NeXus instances by writing a YAML file
which details a hierarchy of data/metadata elements

"""
# -*- coding: utf-8 -*-
#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

# pylint: disable=E1101


import os
import sys
from typing import List

import xml.etree.ElementTree as ET
from xml.dom import minidom
import click

from nexusparser.tools.dataconverter import helpers
from nexusparser.tools.yaml2nxdl import yaml2nxdl_forward_tools
from nexusparser.tools.yaml2nxdl import yaml2nxdl_backward_tools


def pretty_print_xml(xml_root, output_xml):
    """Print better human-readable idented and formatted xml file
using built-in libraries and add preceding XML processing instruction

    """
    dom = minidom.parseString(ET.tostring(
        xml_root, encoding='utf-8', method='xml'))
    sibling = dom.createProcessingInstruction(
        'xml-stylesheet', 'type="text/xsl" href="nxdlformat.xsl"')
    root = dom.firstChild
    dom.insertBefore(sibling, root)
    xml_string = dom.toprettyxml()
    with open(output_xml, "w") as file_out:
        file_out.write(xml_string)


def yaml2nxdl(input_file: str, verbose: bool):
    """Main of the yaml2nxdl converter, creates XML tree,
namespace and schema, then evaluates a dictionary
nest of groups recursively and fields or (their) attributes as childs of the groups

    """
    yml_appdef = yaml2nxdl_forward_tools.yml_reader(input_file)

    if verbose:
        sys.stdout.write(f'input-file: {input_file}\n')
        sys.stdout.write('application/base contains the following root-level entries:\n')
        sys.stdout.write(str(yml_appdef.keys()))
    xml_root = ET.Element(
        'definition', {
            'xmlns': 'http://definition.nexusformat.org/nxdl/3.1',
            'xmlns:xsi': 'http://www.w3.org/2001/XMLSchema-instance',
            'xsi:schemaLocation': 'http://www.w3.org/2001/XMLSchema-instance'
        }
    )
    assert 'category' in yml_appdef.keys(
    ), 'Required root-level keyword category is missing!'
    assert yml_appdef['category'] in ['application', 'base'], 'Only \
application and base are valid categories!'
    assert 'doc' in yml_appdef.keys(), 'Required root-level keyword doc is missing!'

    xml_root.set('extends', 'NXobject')
    xml_root.set('type', 'group')
    if yml_appdef['category'] == 'application':
        xml_root.set('category', 'application')
        del yml_appdef['category']
    else:
        xml_root.set('category', 'base')
        del yml_appdef['category']

    assert isinstance(yml_appdef['doc'], str) and yml_appdef['doc'] != '', 'Doc \
has to be a non-empty string!'
    doctag = ET.SubElement(xml_root, 'doc')
    doctag.text = yml_appdef['doc']
    del yml_appdef['doc']

    if 'symbols' in yml_appdef.keys():
        yaml2nxdl_forward_tools.xml_handle_symbols(xml_root, yml_appdef['symbols'])
        del yml_appdef['symbols']

    assert len(yml_appdef.keys()) == 1, 'Accepting at most keywords: category, \
doc, symbols, and NX... at root-level!'
    keyword = list(yml_appdef.keys())[0]  # which is the only one
    assert (keyword[0:3] == '(NX' and keyword[-1:] == ')' and len(keyword) > 4), 'NX \
keyword has an invalid pattern, or is too short!'
    xml_root.set('name', keyword[1:-1])
    yaml2nxdl_forward_tools.recursive_build(xml_root, yml_appdef[keyword], verbose)

    pretty_print_xml(xml_root, input_file.split(".", 1)[0] + '.nxdl.xml')
    if verbose:
        sys.stdout.write('Parsed YAML to NXDL successfully\n')


def get_node_parent_info(tree, node):
    """Return tuple of (parent, index) where:
        parent = node of parent within tree
        index = index of node under parent"""

    parent_map = {c: p for p in tree.iter() for c in p}
    parent = parent_map[node]
    return parent, list(parent).index(node)


class Nxdl2yaml():
    """Parse XML file and print a YML file

"""

    def __init__(
            self,
            symbol_list: List[str],
            root_level_definition: List[str],
            root_level_doc='',
            root_level_symbols=''):
        self.append_flag = True
        self.found_definition = False
        self.jump_symbol_child = False
        self.root_level_doc = root_level_doc
        self.root_level_symbols = root_level_symbols
        self.root_level_definition = root_level_definition
        self.symbol_list = symbol_list

    def handle_symbols(self, depth, node):
        """Handle symbols field and its childs

        """
        self.root_level_symbols = '{indent}{tag}: {text}'.format(
            indent=0 * '  ',
            tag=helpers.remove_namespace_from_tag(node.tag),
            text=node.text.strip() if node.text else '')
        depth += 1
        for child in list(node):
            tag = helpers.remove_namespace_from_tag(child.tag)
            if tag == ('doc'):
                self.symbol_list.append(
                    '{indent}{tag}: "{text}"'.format(
                        indent=1 * '  ',
                        tag=helpers.remove_namespace_from_tag(child.tag),
                        text=child.text.strip().replace('\"', '\'') if child.text else ''))
            elif tag == ('symbol'):
                if 'doc' in child.attrib:
                    self.symbol_list.append(
                        '{indent}{key}: "{value}"'.format(
                            indent=0 * '  ',
                            key=child.attrib['name'],
                            value=child.attrib['doc'] or ''))
                else:
                    for symbol_doc in list(child):
                        tag = symbol_doc.tag.split("}", 1)[1]
                        if tag == ('doc'):
                            self.symbol_list.append(
                                '{indent}{key}: "{text}"'.format(
                                    indent=1 * '  ',
                                    key=child.attrib['name'],
                                    text=symbol_doc.text.strip().replace('\"', '\'')))

    def handle_definition(self, node):
        """Handle definition group and its attributes

        """
        for item in node.attrib:
            if 'schemaLocation' not in item \
                    and 'name' not in item \
                    and 'extends' not in item \
                    and 'type' not in item:
                self.root_level_definition.append(
                    '{key}: {value}'.format(
                        key=item,
                        value=node.attrib[item] or ''))
        if 'name' in node.attrib.keys():
            self.root_level_definition.append(
                '({value}):'.format(
                    value=node.attrib['name'] or ''))

    def handle_root_level_doc(self, node):
        """Handle the documentation field found at root level

"""
        for child in list(node):
            tag = helpers.remove_namespace_from_tag(child.tag)
            if tag == ('doc'):
                self.root_level_doc = '{indent}{tag}: "{text}"'.format(
                    indent=0 * '  ',
                    tag=helpers.remove_namespace_from_tag(child.tag),
                    text=child.text.strip().replace('\"', '\'') if child.text else '')
                node.remove(child)

    def print_root_level_doc(self, file_out):
        """Print at the root level of YML file \
the general documentation field found in XML file

 """
        file_out.write(
            '{indent}{root_level_doc}\n'.format(
                indent=0 * '  ',
                root_level_doc=self.root_level_doc))
        self.root_level_doc = ''

    def print_root_level_info(self, depth, file_out):
        """Print at the root level of YML file \
the information stored as definition attributes in the XML file

"""
        if depth > 0 \
                and [s for s in self.root_level_definition if "category: application" in s]\
                or depth == 1 \
                and [s for s in self.root_level_definition if "category: base" in s]:
            if self.root_level_symbols:
                file_out.write(
                    '{indent}{root_level_symbols}\n'.format(
                        indent=0 * '  ',
                        root_level_symbols=self.root_level_symbols))
                for symbol in self.symbol_list:
                    file_out.write(
                        '{indent}{symbol}\n'.format(
                            indent=0 * '  ',
                            symbol=symbol))
            if self.root_level_definition:
                for defs in self.root_level_definition:
                    file_out.write(
                        '{indent}{defs}\n'.format(
                            indent=0 * '  ',
                            defs=defs))
            self.found_definition = False

    def recursion_in_xml_tree(self, depth, xml_tree, output_yml, verbose):
        """Descend lower level in xml tree. If we are in the symbols branch, \
the recursive behaviour is not triggered as we already handled the symbols' childs
"""
        tree = xml_tree['tree']
        node = xml_tree['node']
        if self.jump_symbol_child is True:
            self.jump_symbol_child = False
        else:
            for child in list(node):
                xml_tree_children = {'tree': tree, 'node': child}
                Nxdl2yaml.xmlparse(self, output_yml, xml_tree_children, depth, verbose)

    def xmlparse(self, output_yml, xml_tree, depth, verbose):
        """Main of the nxdl2yaml converter.
It parses XML tree,
then prints recursively each level of the tree

    """
        tree = xml_tree['tree']
        node = xml_tree['node']
        if verbose:
            sys.stdout.write(f'Node tag: {helpers.remove_namespace_from_tag(node.tag)}\n')
            sys.stdout.write(f'Attributes: {node.attrib}\n')
        with open(output_yml, "a") as file_out:
            tag = helpers.remove_namespace_from_tag(node.tag)
            if tag == ('definition'):
                self.found_definition = True
                Nxdl2yaml.handle_definition(self, node)
            if depth == 0 and not self.root_level_doc:
                Nxdl2yaml.handle_root_level_doc(self, node)
            if tag == ('doc') and depth != 1:
                parent = get_node_parent_info(tree, node)[0]
                doc_parent = helpers.remove_namespace_from_tag(parent.tag)
                if doc_parent != 'item':
                    yaml2nxdl_backward_tools.handle_not_root_level_doc(depth, node, file_out)
            if tag == ('symbols'):
                Nxdl2yaml.handle_symbols(self, depth, node)
                self.jump_symbol_child = True
            # End of root level definition parsing. Print root-level definitions in file
            if self.root_level_doc \
                    and self.append_flag is True \
                    and (depth in (0, 1)):
                Nxdl2yaml.print_root_level_doc(self, file_out)
            if self.found_definition is True and self.append_flag is True:
                Nxdl2yaml.print_root_level_info(self, depth, file_out)
            # End of print root-level definitions in file
            if tag in ('field', 'group') and depth != 0:
                yaml2nxdl_backward_tools.handle_group_or_field(depth, node, file_out)
            if tag == ('enumeration'):
                yaml2nxdl_backward_tools.handle_enumeration(depth, node, file_out)
            if tag == ('attribute'):
                yaml2nxdl_backward_tools.handle_attributes(depth, node, file_out)
            if tag == ('dimensions'):
                yaml2nxdl_backward_tools.handle_dimension(depth, node, file_out)
        depth += 1
        # Write nested nodes
        Nxdl2yaml.recursion_in_xml_tree(self, depth, xml_tree, output_yml, verbose)


def print_yml(input_file, verbose):
    """Parse an XML file provided as input and print a YML file

"""
    output_yml = input_file[:-9] + '_parsed.yml'
    if os.path.isfile(output_yml):
        os.remove(output_yml)
    my_file = Nxdl2yaml([], [])
    depth = 0
    tree = ET.parse(input_file)
    xml_tree = {'tree': tree.getroot(), 'node': tree.getroot()}
    my_file.xmlparse(output_yml, xml_tree, depth, verbose)


def append_yml(input_file, append, verbose):
    """Append to an existing Nexus base class new elements provided in YML input file \
and print both an XML and YML file of the extended base class.

"""
    nexus_def_path = os.path.join(os.path.abspath(os.path.dirname(__file__)), '../../definitions')
    assert [s for s in os.listdir(os.path.join(nexus_def_path, 'base_classes')
                                  ) if append.strip() == s.split('.nxdl.xml')[0]], \
        'Your base class extension does not match any existing Nexus base classes'
    tree = ET.parse(os.path.join(nexus_def_path + '/base_classes', append + '.nxdl.xml'))
    root = tree.getroot()
    # warning: tmp files are printed on disk and removed at the ends!!
    pretty_print_xml(root, 'tmp.nxdl.xml')
    print_yml('tmp.nxdl.xml', verbose)
    yaml2nxdl('tmp_parsed.yml', verbose)
    tree = ET.parse('tmp_parsed.nxdl.xml')
    tree2 = ET.parse(input_file)
    root_no_duplicates = ET.Element(
        'definition', {'xmlns': 'http://definition.nexusformat.org/nxdl/3.1',
                       'xmlns:xsi': 'http://www.w3.org/2001/XMLSchema-instance',
                       'xsi:schemaLocation': 'http://www.w3.org/2001/XMLSchema-instance'
                       }
    )
    for attribute_keys in root.attrib.keys():
        if attribute_keys != '{http://www.w3.org/2001/XMLSchema-instance}schemaLocation':
            attribute_value = root.attrib[attribute_keys]
            root_no_duplicates.set(attribute_keys, attribute_value)
    for elems in root.iter():
        if 'doc' in elems.tag:
            root_doc = ET.SubElement(root_no_duplicates, 'doc')
            root_doc.text = elems.text
            break
    group = '{http://definition.nexusformat.org/nxdl/3.1}group'
    root_no_duplicates = yaml2nxdl_backward_tools.compare_niac_and_my(tree, tree2, verbose,
                                                                      group,
                                                                      root_no_duplicates)
    field = '{http://definition.nexusformat.org/nxdl/3.1}field'
    root_no_duplicates = yaml2nxdl_backward_tools.compare_niac_and_my(tree, tree2, verbose,
                                                                      field,
                                                                      root_no_duplicates)
    attribute = '{http://definition.nexusformat.org/nxdl/3.1}attribute'
    root_no_duplicates = yaml2nxdl_backward_tools.compare_niac_and_my(tree, tree2, verbose,
                                                                      attribute,
                                                                      root_no_duplicates)
    pretty_print_xml(root_no_duplicates, f'{input_file.split(".", 1)[0]}'
                     f'_appended.nxdl.xml')
    print_yml(input_file.split(".", 1)[0] + '_appended.nxdl.xml', verbose)
    yaml2nxdl(input_file.split(".", 1)[0] + '_appended_parsed.yml', verbose)
    os.rename(f'{input_file.split(".", 1)[0]}_appended_parsed.yml',
              f'{input_file.split(".", 1)[0]}_appended.yml')
    os.rename(f'{input_file.split(".", 1)[0]}_appended_parsed.nxdl.xml',
              f'{input_file.split(".", 1)[0]}_appended.nxdl.xml')
    os.remove('tmp.nxdl.xml')
    os.remove('tmp_parsed.yml')
    os.remove('tmp_parsed.nxdl.xml')


@click.command()
@click.option(
    '--input-file',
    help='The path to the XML or YAML input data file to read and create \
a YAML or XML file from, respectively.'
)
@click.option(
    '--append',
    help='Parse xml file and append to base class, given that the xml file has same name \
of an existing base class'
)
@click.option(
    '--verbose',
    is_flag=True,
    default=False,
    help='Print in standard output keywords and value types to help \
possible issues in yaml files'
)
def launch_tool(input_file, verbose, append):
    """Main function that distiguishes the input file format and launches the tools.

"""
    if input_file.split(".", 1)[1] in ('yml', 'yaml'):
        yaml2nxdl(input_file, verbose)
        if append:
            append_yml(input_file.split(".", 1)[0] + '.nxdl.xml', append, verbose)
        else:
            pass
    elif input_file.split(".", 1)[1] == 'nxdl.xml':
        if not append:
            print_yml(input_file, verbose)
        else:
            append_yml(input_file, append, verbose)


if __name__ == '__main__':
    launch_tool().parse()  # pylint: disable=no-value-for-parameter



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD.
# See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import os
import logging
import numpy as np
from ase.io.trajectory import Trajectory

from nomad.units import ureg
from nomad.parsing.file_parser import FileParser
from nomad.datamodel.metainfo.simulation.run import Run, Program
from nomad.datamodel.metainfo.simulation.method import (
    Method, ForceField, Model)
from nomad.datamodel.metainfo.simulation.system import (
    System, Atoms, Constraint)
from nomad.datamodel.metainfo.simulation.calculation import (
    Calculation, Energy, EnergyEntry, Forces, ForcesEntry)
from nomad.datamodel.metainfo.workflow import (
    Workflow, GeometryOptimization, MolecularDynamics)


class TrajParser(FileParser):
    def __init__(self):
        super().__init__()

    @property
    def traj(self):
        if self._file_handler is None:
            self._file_handler = Trajectory(self.mainfile, 'r')
            # check if traj file is really asap
            if 'calculator' in self._file_handler.backend.keys():
                if self._file_handler.backend.calculator.name != 'emt':  # pylint: disable=E1101
                    self.logger.error('Trajectory is not ASAP.')
                    self._file_handler = None
        return self._file_handler

    def get_version(self):
        if hasattr(self.traj, 'ase_version') and self.traj.ase_version:
            return self.traj.ase_version
        else:
            return '3.x.x'


class AsapParser:
    def __init__(self):
        self.traj_parser = TrajParser()

    def init_parser(self):
        self.traj_parser.mainfile = self.filepath
        self.traj_parser.logger = self.logger

    def parse_system(self, traj):
        sec_system = self.archive.run[0].m_create(System)

        sec_atoms = sec_system.m_create(Atoms)
        sec_atoms.lattice_vectors = traj.get_cell() * ureg.angstrom
        sec_atoms.labels = traj.get_chemical_symbols()
        sec_atoms.positions = traj.get_positions() * ureg.angstrom
        sec_atoms.periodic = traj.get_pbc()
        if traj.get_velocities() is not None:
            sec_atoms.velocities = traj.get_velocities() * (ureg.angstrom / ureg.fs)

        def get_constraint_name(constraint):
            def index():
                d = constraint['kwargs'].get('direction')
                return ((d / np.linalg.norm(d)) ** 2).argsort()[2]

            name = constraint.get('name')
            if name == 'FixedPlane':
                return ['fix_yz', 'fix_xz', 'fix_xy'][index()]
            elif name == 'FixedLine':
                return ['fix_x', 'fix_y', 'fix_z'][index()]
            elif name == 'FixAtoms':
                return 'fix_xyz'
            else:
                return name

        for constraint in traj.constraints:
            sec_constraint = sec_system.m_create(Constraint)
            as_dict = constraint.todict()
            indices = as_dict['kwargs'].get('a', as_dict['kwargs'].get('indices'))
            sec_constraint.indices = np.asarray(indices)
            sec_constraint.kind = get_constraint_name(as_dict)

    def parse_scc(self, traj):
        sec_scc = self.archive.run[0].m_create(Calculation)

        try:
            sec_energy = sec_scc.m_create(Energy)
            sec_energy.total = EnergyEntry(value=traj.get_total_energy() * ureg.eV)
        except Exception:
            pass

        try:
            sec_forces = sec_scc.m_create(Forces)
            sec_forces.total = ForcesEntry(
                value=traj.get_forces() * ureg.eV / ureg.angstrom,
                value_raw=traj.get_forces(apply_constraint=False) * ureg.eV / ureg.angstrom)
        except Exception:
            pass

    def parse_method(self):
        traj = self.traj_parser.traj
        sec_method = self.archive.run[0].m_create(Method)

        if traj[0].calc is not None:
            sec_method.force_field = ForceField(model=[Model(name=traj[0].calc.name)])

        description = traj.description if hasattr(traj, 'description') else dict()
        if not description:
            return

        sec_workflow = self.archive.m_create(Workflow)

        calc_type = description.get('type')
        if calc_type == 'optimization':
            sec_workflow.type = 'geometry_optimization'
            sec_geometry_opt = sec_workflow.m_create(GeometryOptimization)
            sec_geometry_opt.method = description.get('optimizer', '').lower()
            sec_geometry_opt.x_asap_maxstep = description.get('maxstep', 0)
        elif calc_type == 'molecular-dynamics':
            sec_workflow.type = 'molecular_dynamics'
            sec_md = sec_workflow.m_create(MolecularDynamics)
            sec_md.x_asap_timestep = description.get('timestep', 0)
            sec_md.x_asap_temperature = description.get('temperature', 0)

            md_type = description.get('md-type', '')
            if 'Langevin' in md_type:
                sec_md.ensemble_type = 'NVT'
                sec_md.x_asap_langevin_friction = description.get('friction', 0)
            elif 'NVT' in md_type:
                sec_md.ensemble_type = 'NVT'
            elif 'Verlet' in md_type:
                sec_md.ensemble_type = 'NVE'
            elif 'NPT' in md_type:
                sec_md.ensemble_type = 'NPT'

    def parse(self, filepath, archive, logger):
        self.filepath = os.path.abspath(filepath)
        self.archive = archive
        self.maindir = os.path.dirname(self.filepath)
        self.logger = logger if logger is not None else logging

        self.init_parser()

        if self.traj_parser.traj is None:
            return

        sec_run = self.archive.m_create(Run)
        sec_run. program = Program(name='ASAP', version=self.traj_parser.get_version())

        # TODO do we build the topology and method for each frame
        self.parse_method()
        for traj in self.traj_parser.traj:
            self.parse_system(traj)
            self.parse_scc(traj)
            # add references to scc
            sec_scc = sec_run.calculation[-1]
            sec_scc.method_ref = sec_run.method[-1]
            sec_scc.system_ref = sec_run.system[-1]



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD.
# See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import numpy as np            # pylint: disable=unused-import
import typing                 # pylint: disable=unused-import
from nomad.metainfo import (  # pylint: disable=unused-import
    MSection, MCategory, Category, Package, Quantity, Section, SubSection, SectionProxy,
    Reference
)
from nomad.datamodel.metainfo import simulation


m_package = Package()


class Run(simulation.run.Run):

    m_def = Section(validate=False, extends_base_section=True)

    x_fplo_program_version_sub = Quantity(
        type=str,
        shape=[],
        description='''
        FPLO sub version
        ''')

    x_fplo_program_compilation_options = Quantity(
        type=str,
        shape=[],
        description='''
        FPLO compilation options
        ''')


class System(simulation.system.System):

    m_def = Section(validate=False, extends_base_section=True)

    x_fplo_reciprocal_cell = Quantity(
        type=np.dtype(np.float64),
        shape=[3, 3],
        unit='1 / meter',
        description='''
        Reciprocal Lattice vectors (in Cartesian coordinates). The first index runs over
        the $x,y,z$ Cartesian coordinates, and the second index runs over the 3 lattice
        vectors.
        ''')

    x_fplo_atom_idx = Quantity(
        type=np.dtype(np.int32),
        shape=['number_of_atoms'],
        description='''
        FPLO-internal index for each atom
        ''')

    x_fplo_atom_wyckoff_idx = Quantity(
        type=np.dtype(np.int32),
        shape=['number_of_atoms'],
        description='''
        Wyckoff position index of each atom
        ''')

    x_fplo_atom_cpa_block = Quantity(
        type=np.dtype(np.int32),
        shape=['number_of_atoms'],
        description='''
        CPA block of each atom
        ''')

    x_fplo_structure_type = Quantity(
        type=str,
        shape=[],
        description='''
        FPLO structure type: Crystal/Molecule
        ''')


class Method(simulation.method.Method):

    m_def = Section(validate=False, extends_base_section=True)

    x_fplo_xc_functional_number = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        FPLO number xc functional
        ''')

    x_fplo_xc_functional = Quantity(
        type=str,
        shape=[],
        description='''
        FPLO notation of xc functional
        ''')

    x_fplo_dft_plus_u_projection_type = Quantity(
        type=str,
        shape=[],
        description='''
        FPLO notation of DFT+U projection
        ''')

    x_fplo_dft_plus_u_functional = Quantity(
        type=str,
        shape=[],
        description='''
        FPLO notation of DFT+U functional
        ''')


class DFTPlusU(simulation.method.DFTPlusU):

    m_def = Section(validate=False, extends_base_section=True)

    x_fplo_dft_plus_u_orbital_element = Quantity(
        type=str,
        shape=[],
        description='''
        FPLO: Atom/Orbital dependent DFT+U property: element
        ''')

    x_fplo_dft_plus_u_orbital_species = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        FPLO: Atom/Orbital dependent DFT+U property: species index
        ''')

    x_fplo_dft_plus_u_orbital_F0 = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        FPLO: Atom/Orbital dependent DFT+U property: value F0
        ''')

    x_fplo_dft_plus_u_orbital_F2 = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        FPLO: Atom/Orbital dependent DFT+U property: value F2
        ''')

    x_fplo_dft_plus_u_orbital_F4 = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        FPLO: Atom/Orbital dependent DFT+U property: value F4
        ''')

    x_fplo_dft_plus_u_orbital_F6 = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        FPLO: Atom/Orbital dependent DFT+U property: value F6
        ''')



#!/usr/bin/env python3
"""This file collects the function used in the reverse tool nxdl2yaml.

"""
# -*- coding: utf-8 -*-
#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

# pylint: disable=E1101

import sys


def handle_group_or_field(depth, node, file_out):
    """Handle all the possible attributes that come along a field or group

"""
    if "name" in node.attrib and "type" in node.attrib:
        file_out.write(
            '{indent}{name}({value1}):\n'.format(
                indent=depth * '  ',
                name=node.attrib['name'] or '',
                value1=node.attrib['type'] or ''))
    if "name" in node.attrib and "type" not in node.attrib:
        file_out.write(
            '{indent}{name}:\n'.format(
                indent=depth * '  ',
                name=node.attrib['name'] or ''))
    if "name" not in node.attrib and "type" in node.attrib:
        file_out.write(
            '{indent}({type}):\n'.format(
                indent=depth * '  ',
                type=node.attrib['type'] or ''))
    if "minOccurs" in node.attrib and "maxOccurs" in node.attrib:
        file_out.write(
            '{indent}exists: [min, {value1}, max, {value2}]\n'.format(
                indent=(depth + 1) * '  ',
                value1=node.attrib['minOccurs'] or '',
                value2=node.attrib['maxOccurs'] or ''))
    if "minOccurs" in node.attrib \
            and "maxOccurs" not in node.attrib \
            and node.attrib['minOccurs'] == "1":
        file_out.write(
            '{indent}{name}: required \n'.format(
                indent=(depth + 1) * '  ',
                name='exists'))
    if "recommended" in node.attrib and node.attrib['recommended'] == "true":
        file_out.write(
            '{indent}exists: recommended\n'.format(
                indent=(depth + 1) * '  '))
    if "units" in node.attrib:
        file_out.write(
            '{indent}unit: {value}\n'.format(
                indent=(depth + 1) * '  ',
                value=node.attrib['units'] or ''))


def handle_dimension(depth, node, file_out):
    """Handle the dimension field

"""
    file_out.write(
        '{indent}{tag}:\n'.format(
            indent=depth * '  ',
            tag=node.tag.split("}", 1)[1]))
    if 'rank' in node.attrib:
        file_out.write(
            '{indent}rank: {rank}\n'.format(
                indent=(depth + 1) * '  ',
                rank=node.attrib['rank']))
    dim_list = ''
    for child in list(node):
        tag = child.tag.split("}", 1)[1]
        if tag == ('dim'):
            dim_list = dim_list + '[{index}, {value}], '.format(
                index=child.attrib['index'],
                value=child.attrib['value'])
    file_out.write(
        '{indent}dim: [{value}]\n'.format(
            indent=(depth + 1) * '  ',
            value=dim_list[:-2] or ''))


def handle_attributes(depth, node, file_out):
    """Handle the attributes parsed from the xml file

"""
    file_out.write(
        '{indent}{escapesymbol}{key}:\n'.format(
            indent=depth * '  ',
            escapesymbol=r'\@',
            key=node.attrib['name']))


def handle_enumeration(depth, node, file_out):
    """Handle the enumeration field parsed from the xml file.

If the enumeration items contain a doc field, the yaml file will contain items as child
fields of the enumeration field.

If no doc are inherited in the enumeration items, a list of the items is given for the
enumeration list.

"""
    check_doc = []
    for child in list(node):
        if list(child):
            check_doc.append(list(child))
    if check_doc:
        file_out.write(
            '{indent}{tag}: \n'.format(
                indent=depth * '  ',
                tag=node.tag.split("}", 1)[1]))
        for child in list(node):
            tag = child.tag.split("}", 1)[1]
            if tag == ('item'):
                file_out.write(
                    '{indent}{value}: \n'.format(
                        indent=(depth + 1) * '  ',
                        value=child.attrib['value']))
                if list(child):
                    for item_doc in list(child):
                        item_doc_depth = depth + 2
                        handle_not_root_level_doc(item_doc_depth, item_doc, file_out)
    else:
        file_out.write(
            '{indent}{tag}:'.format(
                indent=depth * '  ',
                tag=node.tag.split("}", 1)[1]))
        enum_list = ''
        for child in list(node):
            tag = child.tag.split("}", 1)[1]
            if tag == ('item'):
                enum_list = enum_list + '{value}, '.format(
                    value=child.attrib['value'])
        file_out.write(
            ' [{enum_list}]\n'.format(
                enum_list=enum_list[:-2] or ''))


def handle_not_root_level_doc(depth, node, file_out):
    """Handle docs field along the yaml file

"""
    file_out.write(
        '{indent}{tag}: "{text}"\n'.format(
            indent=depth * '  ',
            tag=node.tag.split("}", 1)[1],
            text=node.text.strip().replace('\"', '\'') if node.text else ''))


def get_node_parent_info(tree, node):
    """Return tuple of (parent, index) where:
        parent = node of parent within tree
        index = index of node under parent"""

    parent_map = {c: p for p in tree.iter() for c in p}
    parent = parent_map[node]
    return parent, list(parent).index(node)


def compare_niac_and_my(tree, tree2, verbose, node, root_no_duplicates):
    """This function creates two trees with Niac XML file and My XML file.
The main aim is to compare the two trees and create a new one that is the
union of the two initial trees.

"""
    root = tree.getroot()
    root2 = tree2.getroot()
    attrs_list_niac = []
    for nodo in root.iter(node):
        attrs_list_niac.append(nodo.attrib)
    if verbose:
        sys.stdout.write('Attributes found in Niac file: \n')
        sys.stdout.write(str(attrs_list_niac) + '\n')
        sys.stdout.write('  \n')
        sys.stdout.write('Started merging of Niac and My file... \n')
    for elem in root.iter(node):
        if verbose:
            sys.stdout.write('- Niac element inserted: \n')
            sys.stdout.write(str(elem.attrib) + '\n')
        index = get_node_parent_info(tree, elem)[1]
        root_no_duplicates.insert(index, elem)

    for elem2 in root2.iter(node):
        index = get_node_parent_info(tree2, elem2)[1]
        if elem2.attrib not in attrs_list_niac:
            if verbose:
                sys.stdout.write('- My element inserted: \n')
                sys.stdout.write(str(elem2.attrib) + '\n')
            root_no_duplicates.insert(index, elem2)

    if verbose:
        sys.stdout.write('     \n')
    return root_no_duplicates



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD.
# See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import numpy as np            # pylint: disable=unused-import
import typing                 # pylint: disable=unused-import
from nomad.metainfo import (  # pylint: disable=unused-import
    MSection, MCategory, Category, Package, Quantity, Section, SubSection, SectionProxy,
    Reference, JSON
)
from nomad.datamodel.metainfo import simulation


m_package = Package()


class Run(simulation.run.Run):

    m_def = Section(validate=False, extends_base_section=True)

    vasp_build_date = Quantity(
        type=str,
        shape=[],
        description='''
        build date as string
        ''')

    vasp_src_date = Quantity(
        type=str,
        shape=[],
        description='''
        date of last modification of the source as string
        ''')


class Method(simulation.method.Method):

    m_def = Section(validate=False, extends_base_section=True)

    x_vasp_incar_in = Quantity(
        type=JSON,
        shape=[],
        description='''
        contains all the user-input INCAR parameters
        ''')

    x_vasp_incar_out = Quantity(
        type=JSON,
        shape=[],
        description='''
        contains the actual INCAR parameters used by VASP at runtime
        ''')

    x_vasp_unknown_incars = Quantity(
        type=JSON,
        shape=[],
        description='''
        INCAR variables uknown wrt to Vasp Wiki
        ''')

    x_vasp_atom_kind_refs = Quantity(
        type=simulation.method.AtomParameters,
        shape=['number_of_atoms'],
        description='''
        reference to the atom kinds of each atom
        ''')

    x_vasp_k_points_generation_method = Quantity(
        type=str,
        shape=[],
        description='''
        k points generation  method
        ''')

    x_vasp_k_points = Quantity(
        type=int,
        shape=['x_vasp_number_of_k_points'],
        description='''
        k points
        ''')

    x_vasp_numer_of_magmom = Quantity(
        type=int,
        shape=[],
        description='''
        number of magnetic moments, number_of_atoms for ISPIN = 2, 3*number of atoms for
        non-collinear magnetic systems
        ''')

    x_vasp_tetrahedrons_list = Quantity(
        type=np.dtype(np.int32),
        shape=['N', 5],
        description='''
        Rows of 5 elements. First the weight (symmetry degeneration), then the four corner
        points of each tetrahedron.
        ''')

    x_vasp_tetrahedron_volume = Quantity(
        type=np.dtype(np.float64),
        shape=[1],
        description='''
        Volume weight of a single tetrahedron (all tetra's must have the same volume)
        ''')

    x_vasp_nose_thermostat = Quantity(
        type=np.dtype(np.float64),
        shape=[4],
        description='''
        Nose thermostat output
        ''')


class System(simulation.system.System):

    m_def = Section(validate=False, extends_base_section=True)

    x_vasp_selective_dynamics = Quantity(
        type=np.dtype(bool),
        shape=['number_of_atoms', 3],
        description='''
        Boolean array to eiter allow or forbid coordinate modifications during relaxation
        ''')



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""The writer class for writing a Nexus file in accordance with a given NXDL."""

import copy
import logging
import sys
import xml.etree.ElementTree as ET

import h5py
import numpy as np

from nexusparser.tools.dataconverter import helpers
from nexusparser.tools import nexus

logger = logging.getLogger(__name__)  # pylint: disable=C0103
logger.setLevel(logging.INFO)
logger.addHandler(logging.StreamHandler(sys.stdout))


def does_path_exist(path, h5py_obj) -> bool:
    """Returns true if the requested path exists in the given h5py object."""
    try:
        h5py_obj[helpers.convert_data_dict_path_to_hdf5_path(path)]  # pylint: disable=W0106
        return True
    except KeyError:
        return False


def is_not_data_empty(value) -> bool:
    """Returns True if value is an Numpy array or not None."""
    if isinstance(value, np.ndarray) or value is not None:
        return True
    return False


def get_namespace(element) -> str:
    """Extracts the namespace for elements in the NXDL"""
    return element.tag[element.tag.index("{"):element.tag.rindex("}") + 1]


class Writer:
    """The writer class for writing a Nexus file in accordance with a given NXDL.

    Args:
        data (dict): Dictionary containing the data to convert.
        nxdl_path (str): Path to the nxdl file to use during conversion.
        output_path (str): Path to the output Nexus file.

    Attributes:
        data (dict): Dictionary containing the data to convert.
        nxdl_path (str): Path to the nxdl file to use during conversion.
        output_path (str): Path to the output Nexus file.
        output_nexus (h5py.File): The h5py file object to manipulate output file.
        nxdl_data (dict): Stores xml data from given nxdl file to use during conversion.
        nxs_namespace (str): The namespace used in the NXDL tags. Helps search for XML children.
    """

    def __init__(self, data: dict = None, nxdl_path: str = None, output_path: str = None):
        """Constructs the necessary objects required by the Writer class."""
        self.data = data
        self.nxdl_path = nxdl_path
        self.output_path = output_path
        self.output_nexus = h5py.File(self.output_path, "w")
        self.nxdl_data = ET.parse(self.nxdl_path).getroot()
        self.nxs_namespace = get_namespace(self.nxdl_data)

    def __nxdl_to_attrs(self, path: str = '/') -> dict:
        """
        Return a dictionary of all the attributes at the given path in the NXDL and
        the required attribute values that were requested in the NXDL from the data.
        """
        nxdl_path = helpers.convert_data_converter_dict_to_nxdl_path(path)
        elem = copy.deepcopy(nexus.get_node_at_nxdl_path(nxdl_path, elem=self.nxdl_data))
        if elem is None:
            raise Exception(f"Attributes were not found for {path}. "
                            "Please check this entry in the template dictionary.")

        # Remove the name attribute as we only use it to name the HDF5 entry
        if "name" in elem.attrib.keys():
            del elem.attrib["name"]

        # Fetch values for required attributes requested by the NXDL
        for attr_name in elem.findall(f"{self.nxs_namespace}attribute"):
            elem.attrib[attr_name.get('name')] = self.data[f"{path}/@{attr_name.get('name')}"] or ''

        return elem.attrib

    def ensure_and_get_parent_node(self, path: str, undocumented_paths) -> h5py.Group:
        """Returns the parent if it exists for a given path else creates the parent group."""
        parent_path = path[0:path.rindex('/')] or '/'
        parent_path_hdf5 = helpers.convert_data_dict_path_to_hdf5_path(parent_path)
        parent_undocumented_paths = [path[0:path.rindex("/")] or "/" for path in undocumented_paths]
        if not does_path_exist(parent_path, self.output_nexus):
            parent = self.ensure_and_get_parent_node(parent_path, parent_undocumented_paths)
            grp = parent.create_group(parent_path_hdf5)
            if path not in undocumented_paths:
                attrs = self.__nxdl_to_attrs(parent_path)
                if attrs is not None:
                    grp.attrs['NX_class'] = attrs["type"]
            return grp
        return self.output_nexus[parent_path_hdf5]

    def write(self):
        """Writes the Nexus file with previously validated data from the reader with NXDL attrs."""
        for path, value in self.data.items():
            try:
                if path[path.rindex('/') + 1:] == '@units':
                    continue

                entry_name = helpers.get_name_from_data_dict_entry(path[path.rindex('/') + 1:])

                data = value if is_not_data_empty(value) else "NOT_PROVIDED"

                if entry_name[0] != "@":
                    grp = self.ensure_and_get_parent_node(path, self.data.undocumented.keys())

                    if isinstance(data, dict) and 'link' in data.keys():
                        where_to_place_link = helpers.convert_data_dict_path_to_hdf5_path(path)
                        self.output_nexus[where_to_place_link] = h5py.SoftLink(
                            helpers.convert_data_dict_path_to_hdf5_path(data['link']))

                    else:
                        dataset = grp.create_dataset(entry_name, data=data)
                        units_key = f"{path}/@units"

                        if units_key in self.data.keys() and self.data[units_key] is not None:
                            dataset.attrs["units"] = self.data[units_key]
                        else:
                            dataset.attrs["units"] = "NOT_PROVIDED"
                else:
                    dataset = self.ensure_and_get_parent_node(path, self.data.undocumented.keys())
                    dataset.attrs[entry_name[1:]] = data
            except Exception as exception:
                raise Exception(f"Unkown error occured writing the path: {path} "
                                f"with the following message: {str(exception)}")

        self.output_nexus.close()
